/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: Graph.cpp                                                                  *
 * @author: Xu Le, He Jianfan                                                        *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#include "Graph.h"

extern int log_level;

extern uint32_t networkNodeNum;
extern uint32_t linkNum;
extern uint32_t consumptionNodeNum;
extern uint32_t serverDeploymentFee;

/**
 * @brief maintaining the map from arc ID to the point to arc, cost only O(1).
 *
 * 1. range of links that between network nodes is [0, 4*linkNum-1], and all integer indices are filled;
 * 2. range of links that between access network node and consumption node is
 *    [4*linkNum, 4*linkNum+4*consumptionNodeNum-2], and all even indices are filled;
 * 3. range of links that between consumption nodes and super sink node is
 *    [4*linkNum+4*consumptionNodeNum, 4*linkNum+8*consumptionNodeNum-2], and all even indices are filled;
 * 4. range of links that between super source node and server deployment nodes is
 *    [4*linkNum+1, 4*linkNum+4*sourceNum-1], and all odd indices are filled.
 */
static std::vector<struct Arc*> arcTable;
/** maintaining the map from comsumption node ID to its access network node ID, cost only O(1). */
std::vector<int> accessTable;
/** maintaining the map from comsumption node ID to its bandwidth demand, cost only O(1). */
std::vector<int> demandTable;

/** automatically increase ID for all arcs. */
static int arcIDIncrement = 0;

struct Arc* oppositeArc(int ownID) // ownID must less than 4*linkNum
{
	if (ownID % 4 == 0)
		return arcTable[ownID+1]; // 0 to 1
	else
		return arcTable[ownID-1]; // 1 to 0
}

struct Arc* oppositeReverseArc(int ownID) // ownID must less than 4*linkNum
{
	if (ownID % 4 == 0)
		return arcTable[ownID+3]; // 0 to 3
	else
		return arcTable[ownID+1]; // 1 to 2
}

Arc::Arc() : arcID(arcIDIncrement), srcID(0), dstID(0), bandwidth(0), rental(INF), initialCap(0), initialCost(0), nextArc(NULL), reverseArc(NULL)
{
	arcIDIncrement += 2;
}

/**
 * constructor of class Graph.
 */
Graph::Graph(uint32_t _node_num) : nodeNum(_node_num), useZKW(false), exchangeOperation(false), cost(0), sumFlux(0)
{
	NodeList = new Node[_node_num + 2]; // plus two, one for super source, one for super sink
	color = new Colortype[_node_num + 2];
	dist = new int[_node_num + 2];
	f = new int[_node_num + 2];
	pi = new int[_node_num + 2];
	arcRecord = new Arc*[_node_num + 2];
}

/**
 * destructor of class Graph.
 */
Graph::~Graph()
{
	delete []arcRecord;
	delete []pi;
	delete []dist;
	delete []f;
	delete []color;
	delete []NodeList;
}

/**
 * construct the graph in the structure of adjacency list.
 */
void Graph::construct(std::list<struct LinkTuple>& linkTuples)
{
	std::list<struct LinkTuple>::iterator iter1 = linkTuples.begin(), iter2;
	std::vector<bool> arcIsAdded(nodeNum, false);
	uint32_t curNode = 0;
	uint32_t i = 0, rank = 2;
	struct Arc *firstarc = NULL, *parc = NULL, *qarc = NULL, *rarc = NULL;

	uint32_t arcTableSizeDecider = 4 * std::max(networkNodeNum, 2*consumptionNodeNum);
	arcTable.resize(4*linkNum + arcTableSizeDecider);

	firstarc = new Arc;
	firstarc->srcID = iter1->src;
	firstarc->dstID = iter1->dst;
	firstarc->bandwidth = iter1->bandwidth;
	firstarc->rental = iter1->rental;
	firstarc->initialCap = iter1->bandwidth;
	firstarc->initialCost = iter1->rental;
	firstarc->nextArc = NULL;
	rarc = new Arc;
	rarc->srcID = firstarc->dstID;
	rarc->dstID = firstarc->srcID;
	rarc->rental = -firstarc->rental;
	rarc->initialCost = rarc->rental;
	firstarc->reverseArc = rarc;
	rarc->reverseArc = firstarc;

	curNode = iter1->src; // node index start with 0
	arcIsAdded[curNode] = true;
	NodeList[curNode].ID = firstarc->srcID;
	NodeList[curNode].firstArc = firstarc;
	NodeList[curNode].lastArc = firstarc;
	arcTable[firstarc->arcID] = firstarc;
	arcTable[rarc->arcID] = rarc;
	while (true)
	{
		iter2 = iter1; // iter2 is always a forword element of iter1
		++iter2;
		if (iter2 == linkTuples.end())
			break;
		if (iter2->src == iter1->src)
		{
			if (rank % 2 == 0)
			{
				parc = new Arc;
				parc->srcID = iter2->src;
				parc->dstID = iter2->dst;
				parc->bandwidth = iter2->bandwidth;
				parc->rental = iter2->rental;
				parc->initialCap = iter2->bandwidth;
				parc->initialCost = iter2->rental;
				parc->nextArc = NULL;
				rarc = new Arc;
				rarc->srcID = parc->dstID;
				rarc->dstID = parc->srcID;
				rarc->rental = -parc->rental;
				rarc->initialCost = rarc->rental;
				parc->reverseArc = rarc;
				rarc->reverseArc = parc;

				if (rank == 2)
					firstarc->nextArc = parc;
				else
					qarc->nextArc = parc;
				NodeList[curNode].lastArc = parc;
				arcTable[parc->arcID] = parc;
				arcTable[rarc->arcID] = rarc;
			}
			else
			{
				qarc = new Arc;
				qarc->srcID = iter2->src;
				qarc->dstID = iter2->dst;
				qarc->bandwidth = iter2->bandwidth;
				qarc->rental = iter2->rental;
				qarc->initialCap = iter2->bandwidth;
				qarc->initialCost = iter2->rental;
				qarc->nextArc = NULL;
				rarc = new Arc;
				rarc->srcID = qarc->dstID;
				rarc->dstID = qarc->srcID;
				rarc->rental = -qarc->rental;
				rarc->initialCost = rarc->rental;
				qarc->reverseArc = rarc;
				rarc->reverseArc = qarc;

				parc->nextArc = qarc;
				NodeList[curNode].lastArc = qarc;
				arcTable[qarc->arcID] = qarc;
				arcTable[rarc->arcID] = rarc;
			}
			rank++;
		}
		else
		{
			rank = 2;
			curNode = iter2->src;
			arcIsAdded[curNode] = true;

			firstarc = new Arc;
			firstarc->srcID = curNode;
			firstarc->dstID = iter2->dst;
			firstarc->bandwidth = iter2->bandwidth;
			firstarc->rental = iter2->rental;
			firstarc->initialCap = iter2->bandwidth;
			firstarc->initialCost = iter2->rental;
			firstarc->nextArc = NULL;
			rarc = new Arc;
			rarc->srcID = firstarc->dstID;
			rarc->dstID = firstarc->srcID;
			rarc->rental = -firstarc->rental;
			rarc->initialCost = rarc->rental;
			firstarc->reverseArc = rarc;
			rarc->reverseArc = firstarc;

			NodeList[curNode].ID = curNode;
			NodeList[curNode].firstArc = firstarc;
			NodeList[curNode].lastArc = firstarc;
			arcTable[firstarc->arcID] = firstarc;
			arcTable[rarc->arcID] = rarc;
		}
		++iter1;
	}
	for (i = 0; i < consumptionNodeNum; ++i)
	{
		curNode = accessTable[i];          // access network node
		arcIsAdded[curNode] = true;

		parc = new Arc;
		parc->srcID = curNode;
		parc->dstID = networkNodeNum + i;  // consumption node index
		parc->bandwidth = demandTable[i];  // demand bandwidth of each consumption node
		parc->rental = 0;                  // access link need no rental fee
		parc->initialCap = demandTable[i];
		parc->initialCost = 0;
		parc->nextArc = NULL;
		rarc = new Arc;
		rarc->srcID = parc->dstID;
		rarc->dstID = parc->srcID;
		rarc->rental = 0;
		parc->reverseArc = rarc;
		rarc->reverseArc = parc;

		NodeList[curNode].ID = curNode;
		if (NodeList[curNode].firstArc == NULL)
			NodeList[curNode].firstArc = parc;
		else
			NodeList[curNode].lastArc->nextArc = parc; // append to the last
		NodeList[curNode].lastArc = parc; // update lastArc
		arcTable[parc->arcID] = parc;
		arcTable[rarc->arcID] = rarc;
	}
	for (i = 0; i < nodeNum; ++i)
		if (arcIsAdded[i] == false)
			NodeList[i].ID = i;
}

/**
 * @brief tranform directed graph to be undirected graph.
 */
void Graph::undirectify(std::list<struct LinkTuple>& linkTuples)
{
	std::list<struct LinkTuple>::iterator iter = linkTuples.begin();
	struct Arc *parc = NULL, *rarc = NULL;
	uint32_t curNode = 0;
	
	arcIDIncrement = 1; // set arc x and x+1 be the same link with opposite direction, x is an even integer
	
	for (; iter != linkTuples.end(); ++iter)
	{
		curNode = iter->dst; // alias

		parc = new Arc;
		parc->srcID = curNode;
		parc->dstID = iter->src;
		parc->bandwidth = iter->bandwidth;
		parc->rental = iter->rental;
		parc->initialCap = iter->bandwidth;
		parc->initialCost = iter->rental;
		parc->nextArc = NULL;
		rarc = new Arc;
		rarc->srcID = parc->dstID;
		rarc->dstID = parc->srcID;
		rarc->rental = -parc->rental;
		rarc->initialCost = rarc->rental;
		parc->reverseArc = rarc;
		rarc->reverseArc = parc;

		if (NodeList[curNode].firstArc == NULL)
			NodeList[curNode].firstArc = parc;
		else
			NodeList[curNode].lastArc->nextArc = parc; // append to the last
		NodeList[curNode].lastArc = parc; // update lastArc
		
		arcTable[parc->arcID] = parc;
		arcTable[rarc->arcID] = rarc;
	}
}

/**
 * destruct the graph and free memory.
 */
void Graph::destruct()
{
	for (uint32_t i = 0; i < nodeNum; ++i)
	{
		uint32_t rank = 0;
		struct Arc *freearc1 = NodeList[i].firstArc, *freearc2 = (freearc1 != NULL) ? freearc1->nextArc : NULL;

		while (freearc1 != NULL && freearc2 != NULL)
		{
			if (rank % 2 == 0)
			{
				delete freearc1->reverseArc;
				delete freearc1;
				freearc1 = freearc2->nextArc;
			}
			else
			{
				delete freearc2->reverseArc;
				delete freearc2;
				freearc2 = freearc1->nextArc;
			}
			++rank;
		}
		if (freearc1 != NULL)
		{
			delete freearc1->reverseArc;
			delete freearc1;
		}
		if (freearc2 != NULL)
		{
			delete freearc2->reverseArc;
			delete freearc2;
		}
	}
}

/**
 * display the graph whose structure is adjacency list.
 */
void Graph::display()
{
	if (log_level < DEBUG_LEVEL)
		return;

	printf("Graph's adjacency list displays: \n");
	for (uint32_t i = 0; i < nodeNum; i++)
	{
		struct Arc *darc;
		printf("node: [%d]", NodeList[i].ID);
		darc = NodeList[i].firstArc;
		while (darc != NULL)
		{
			printf(" -> %d(%d,%d)", darc->dstID, darc->bandwidth, darc->rental);
			darc = darc->nextArc;
		}
		printf("\n");
	}
}

/** are you prefer ZKW algorithm to SPFA algorithm? */
void Graph::preferZKW(bool _useZKW)
{
	useZKW = _useZKW;
}

/** a collection of things to be done before executing minimum cost maximum flow algorithm, this function called only once. */
void Graph::preMinimumCostMaximumFlow()
{
	superSourceID = static_cast<int>(networkNodeNum + consumptionNodeNum);
	superSinkID = static_cast<int>(networkNodeNum + consumptionNodeNum + 1);
	NodeList[superSourceID].ID = superSourceID;
	NodeList[superSinkID].ID = superSinkID;
	for (uint32_t l = 0; l < consumptionNodeNum; ++l)
		sumFlux += demandTable[l];

	if (useZKW)
	{
		visited = new bool[nodeNum+2];
		fluxFeeArray = new FluxFee[networkNodeNum*networkNodeNum];
		feeMatrix = new FluxFee*[networkNodeNum]; // a point array, convenient to use form feeMatrix[i][j]
		for (uint32_t k = 0; k < networkNodeNum; ++k)
			feeMatrix[k] = fluxFeeArray + k*networkNodeNum;
		for (int i = 0; i < static_cast<int>(networkNodeNum); ++i)
		{
			for (int j = 0; j < static_cast<int>(networkNodeNum); ++j)
			{
				feeMatrix[i][j].from = i;
				feeMatrix[i][j].to = j;
			}
		}
	}

	_addSuperSink();
}

/** API function for Greedy. */
void Graph::setExchangeOperation(bool _exchange_operation)
{
	exchangeOperation = _exchange_operation;
}

/** decide which concrete minimum cost maximum flow algorithm to be executed by bool variable useZKW. */
long Graph::minimumCostMaximumFlow(std::list<int>& servers)
{
	if (useZKW)
		return _ZKW();
	else
		return _SPFA();
}

/** a minimum cost maximum flow algorithm using priority queue. */
long Graph::_SPFA()
{
	long totalCost = 0;
	int sumFlow = sumFlux;
	int u, v;
	uint32_t totalNodeNum = nodeNum + 2;
	struct Arc *parc = NULL, *rarc = NULL;
	std::vector<int> h(totalNodeNum, 0);
	std::priority_queue<std::pair<int, int>, std::vector<std::pair<int, int> >, std::greater<std::pair<int, int> > > priorityQ;

	while (sumFlow > 0)
	{
		std::fill(dist, dist + totalNodeNum, INF);
		dist[superSourceID] = 0;
		priorityQ.push(std::pair<int, int>(0, superSourceID));
		while (!priorityQ.empty())
		{
			std::pair<int, int> qTop = priorityQ.top();
			priorityQ.pop();
			u = qTop.second;
			if (dist[u] < qTop.first)
				continue;

			parc = NodeList[u].firstArc;
			while (parc != NULL)
			{
				if (parc->arcID < 4*static_cast<int>(linkNum)) // only arcs that between network nodes have opposite arc
				{
					rarc = oppositeReverseArc(parc->arcID);
					if (parc->arcID % 2 == 0) // itself first, reverse arc of its opposite arc second
					{
						v = parc->dstID;
						if (parc->initialCap > 0 && dist[v] > dist[u] + parc->rental + h[u] - h[v])
						{
							dist[v] = dist[u] + parc->rental + h[u] - h[v];
							pi[v] = u;
							arcRecord[v] = parc;
							priorityQ.push(std::pair<int, int>(dist[v], v));
						}
						
						v = rarc->dstID;
						if (rarc->initialCap > 0 && dist[v] > dist[u] + rarc->rental + h[u] - h[v])
						{
							dist[v] = dist[u] + rarc->rental + h[u] - h[v];
							pi[v] = u;
							arcRecord[v] = rarc;
							priorityQ.push(std::pair<int, int>(dist[v], v));
						}
					}
					else // reverse arc of its opposite arc first, itself second
					{
						v = rarc->dstID;
						if (rarc->initialCap > 0 && dist[v] > dist[u] + rarc->rental + h[u] - h[v])
						{
							dist[v] = dist[u] + rarc->rental + h[u] - h[v];
							pi[v] = u;
							arcRecord[v] = rarc;
							priorityQ.push(std::pair<int, int>(dist[v], v));
						}
						
						v = parc->dstID;
						if (parc->initialCap > 0 && dist[v] > dist[u] + parc->rental + h[u] - h[v])
						{
							dist[v] = dist[u] + parc->rental + h[u] - h[v];
							pi[v] = u;
							arcRecord[v] = parc;
							priorityQ.push(std::pair<int, int>(dist[v], v));
						}
					}
				}
				else
				{
					if (u >= static_cast<int>(networkNodeNum) && u < static_cast<int>(networkNodeNum + consumptionNodeNum)) // u is a consumption node
					{
						rarc = NodeList[accessTable[u-(int)networkNodeNum]].firstArc;
						while (rarc != NULL)
						{
							if (rarc->dstID == u)
								break;
							rarc = rarc->nextArc;
						}
						rarc = rarc->reverseArc;

						v = rarc->dstID;
						if (rarc->initialCap > 0 && dist[v] > dist[u] + rarc->rental + h[u] - h[v])
						{
							dist[v] = dist[u] + rarc->rental + h[u] - h[v];
							pi[v] = u;
							arcRecord[v] = rarc;
							priorityQ.push(std::pair<int, int>(dist[v], v));
						}
					}

					v = parc->dstID;
					if (parc->initialCap > 0 && dist[v] > dist[u] + parc->rental + h[u] - h[v])
					{
						dist[v] = dist[u] + parc->rental + h[u] - h[v];
						pi[v] = u;
						arcRecord[v] = parc;
						priorityQ.push(std::pair<int, int>(dist[v], v));
					}
				}
				parc = parc->nextArc;
			}
		}
		if (dist[superSinkID] == INF)
			return -1;
		for (u = 0; u < (int)totalNodeNum; ++u)
			h[u] += dist[u];

		int d = sumFlow;
		for (u = superSinkID; u != superSourceID; u = pi[u])
		{
			parc = arcRecord[u];
			d = std::min(d, parc->initialCap);
		}
		for (u = superSinkID; u != superSourceID; u = pi[u])
		{
			parc = arcRecord[u];
			totalCost += d*parc->rental;
		}
		sumFlow -= d;
		for (u = superSinkID; u != superSourceID; u = pi[u])
		{
			parc = arcRecord[u];
			parc->initialCap -= d;
			parc->reverseArc->initialCap += d;
		}
	}
	
	return totalCost;
}

/** https://artofproblemsolving.com/community/c1368h1020435 */
long Graph::_ZKW()
{
	long totalCost = 0;

	int i = 0, j = 0;
	int augFlow = 0;
	int sumFlow = sumFlux;
	cost = 0;

	for (i = 0; i < static_cast<int>(networkNodeNum); ++i)
		for (j = 0; j < static_cast<int>(networkNodeNum); ++j)
			feeMatrix[i][j].fee = 0;

	do {
		do {
			memset(visited, false, (nodeNum+2) * sizeof(bool));
			augFlow = _augment(superSourceID, INF, superSourceID, totalCost);
			sumFlow -= augFlow;
		}
		while ( augFlow );
	}
	while ( _modifyLabel() );

	if (exchangeOperation)
	{
		while (!fluxFeeQueue.empty())
			fluxFeeQueue.pop();

		for (i = 0; i < static_cast<int>(networkNodeNum); ++i)
			for (j = 0; j < static_cast<int>(networkNodeNum); ++j)
				if (feeMatrix[i][j].fee > 0)
					fluxFeeQueue.push(feeMatrix[i][j]);
	}

	return sumFlow > 0 ? -1 : totalCost;
}

/** find a augment path in graph, this function is called by _ZKW(). */
int Graph::_augment(int u, int f, int from, long& totalCost)
{
	if (u == superSinkID)
	{
		totalCost += cost * f;
		// debug_log("totalCost: %ld, cost: %ld, f: %d\n", totalCost, cost, f);
		return f;
	}

	visited[u] = true;

	int left = f;
	int v = -1;
	struct Arc *parc = NodeList[u].firstArc, *rarc = NULL;
	while (parc != NULL)
	{
		if (parc->arcID < 4*static_cast<int>(linkNum)) // only arcs that between network nodes have opposite arc
		{
			rarc = oppositeReverseArc(parc->arcID);
			if (parc->arcID % 2 == 0) // itself first, reverse arc of its opposite arc second
			{
				v = parc->dstID;
				if (parc->initialCap > 0 && !parc->initialCost && visited[v] == false)
				{
					if (u == superSourceID)
						from = v;
					int delta = _augment(v, left < parc->initialCap ? left : parc->initialCap, from, totalCost);
					if (v == superSinkID)
						feeMatrix[from][accessTable[u-(int)networkNodeNum]].fee += cost * (left < parc->initialCap ? left : parc->initialCap);
					parc->initialCap -= delta;
					parc->reverseArc->initialCap += delta;
					left -= delta;
					if (left == 0)
						return f;
				}

				v = rarc->dstID;
				if (rarc->initialCap > 0 && !rarc->initialCost && visited[v] == false)
				{
					if (u == superSourceID)
						from = v;
					int delta = _augment(v, left < rarc->initialCap ? left : rarc->initialCap, from, totalCost);
					if (v == superSinkID)
						feeMatrix[from][accessTable[u-(int)networkNodeNum]].fee += cost * (left < parc->initialCap ? left : parc->initialCap);
					rarc->initialCap -= delta;
					rarc->reverseArc->initialCap += delta;
					left -= delta;
					if (left == 0)
						return f;
				}
			}
			else // reverse arc of its opposite arc first, itself second
			{
				v = rarc->dstID;
				if (rarc->initialCap > 0 && !rarc->initialCost && visited[v] == false)
				{
					if (u == superSourceID)
						from = v;
					int delta = _augment(v, left < rarc->initialCap ? left : rarc->initialCap, from, totalCost);
					if (v == superSinkID)
						feeMatrix[from][accessTable[u-(int)networkNodeNum]].fee += cost * (left < parc->initialCap ? left : parc->initialCap);
					rarc->initialCap -= delta;
					rarc->reverseArc->initialCap += delta;
					left -= delta;
					if (left == 0)
						return f;
				}

				v = parc->dstID;
				if (parc->initialCap > 0 && !parc->initialCost && visited[v] == false)
				{
					if (u == superSourceID)
						from = v;
					int delta = _augment(v, left < parc->initialCap ? left : parc->initialCap, from, totalCost);
					if (v == superSinkID)
						feeMatrix[from][accessTable[u-(int)networkNodeNum]].fee += cost * (left < parc->initialCap ? left : parc->initialCap);
					parc->initialCap -= delta;
					parc->reverseArc->initialCap += delta;
					left -= delta;
					if (left == 0)
						return f;
				}
			}
		}
		else
		{
			v = parc->dstID;
			if (parc->initialCap > 0 && !parc->initialCost && visited[v] == false)
			{
				if (u == superSourceID)
						from = v;
				int delta = _augment(v, left < parc->initialCap ? left : parc->initialCap, from, totalCost);
				if (v == superSinkID)
					feeMatrix[from][accessTable[u-(int)networkNodeNum]].fee += cost * (left < parc->initialCap ? left : parc->initialCap);
				parc->initialCap -= delta;
				parc->reverseArc->initialCap += delta;
				left -= delta;
				if (left == 0)
					return f;
			}
		}
		parc = parc->nextArc;
	}

	return f - left;
}

/** modify arcs' status in graph, this function is called by _ZKW(). */
bool Graph::_modifyLabel()
{
	int delta = INF;

	uint32_t u = 0;
	int v = 0;
	struct Arc *parc = NULL, *rarc = NULL;

	for (u = 0; u < nodeNum + 2; ++u)
	{
		if (visited[u])
		{
			for (parc = NodeList[u].firstArc; parc != NULL; parc = parc->nextArc)
			{
				if (parc->arcID < 4*static_cast<int>(linkNum)) // only arcs that between network nodes have opposite arc
				{
					rarc = oppositeReverseArc(parc->arcID);
					if (parc->arcID % 2 == 0) // itself first, reverse arc of its opposite arc second
					{
						v = parc->dstID;
						if (parc->initialCap > 0 && visited[v] == false && delta > parc->initialCost)
							delta = parc->initialCost;

						v = rarc->dstID;
						if (rarc->initialCap > 0 && visited[v] == false && delta > rarc->initialCost)
							delta = rarc->initialCost;
					}
					else // reverse arc of its opposite arc first, itself second
					{
						v = rarc->dstID;
						if (rarc->initialCap > 0 && visited[v] == false && delta > rarc->initialCost)
							delta = rarc->initialCost;

						v = parc->dstID;
						if (parc->initialCap > 0 && visited[v] == false && delta > parc->initialCost)
							delta = parc->initialCost;
					}
				}
				else
				{
					v = parc->dstID;
					if (parc->initialCap > 0 && visited[v] == false && delta > parc->initialCost)
						delta = parc->initialCost;
				}
			}
		}
	}

	// info_log("delta: %d\n", delta);
	if (delta == INF)
		return false;

	for (u = 0; u < nodeNum + 2; ++u)
	{
		if (visited[u])
		{
			for (parc = NodeList[u].firstArc; parc != NULL; parc = parc->nextArc)
			{
				if (parc->arcID < 4*static_cast<int>(linkNum)) // only arcs that between network nodes have opposite arc
				{
					rarc = oppositeReverseArc(parc->arcID);
					if (parc->arcID % 2 == 0) // itself first, reverse arc of its opposite arc second
					{
						parc->initialCost -= delta;
						parc->reverseArc->initialCost += delta;

						rarc->initialCost -= delta;
						rarc->reverseArc->initialCost += delta;
					}
					else // reverse arc of its opposite arc first, itself second
					{
						rarc->initialCost -= delta;
						rarc->reverseArc->initialCost += delta;

						parc->initialCost -= delta;
						parc->reverseArc->initialCost += delta;
					}
				}
				else
				{
					parc->initialCost -= delta;
					parc->reverseArc->initialCost += delta;
				}
			}
		}
	}

	if (delta > 0)
		cost += delta; // TODO: handle the situation when delta < 0, try to avoid this happened.

	return true;
}

/** a collection of things to be done after executing minimum cost maximum flow algorithm, this function called only once. */
void Graph::postMinimumCostMaximumFlow()
{
	_delSuperSink();

	if (useZKW)
	{
		delete []visited;
		delete []feeMatrix;
		delete []fluxFeeArray;
	}
}

/** revert the status of all arcs to initialization, this function should be called every time right after algorithm finished. */
void Graph::revertStatusToInitialization()
{
	_revertCapacityToInitialization();

	if (useZKW)
		_revertCostToInitialization();
}

/** revert the capacity of all arcs to initialization, this function is called by revertStatusToInitialization(). */
void Graph::_revertCapacityToInitialization()
{
	uint32_t i;
	for (i = 0; i < 4*linkNum; ++i)
		arcTable[i]->initialCap = arcTable[i]->bandwidth;
	for (i = 4*linkNum; i < 4*linkNum+8*consumptionNodeNum; i += 2)
		arcTable[i]->initialCap = arcTable[i]->bandwidth;
	for (i = 4*linkNum + 1; i < 4*linkNum+4*sourceNum; i += 2)
		arcTable[i]->initialCap = arcTable[i]->bandwidth;
}

/** revert the cost of all arcs to initialization, this function is called by revertStatusToInitialization(). */
void Graph::_revertCostToInitialization()
{
	uint32_t i;
	for (i = 0; i < 4*linkNum; ++i)
		arcTable[i]->initialCost = arcTable[i]->rental;
	for (i = 4*linkNum; i < 4*linkNum+8*consumptionNodeNum; i += 2)
		arcTable[i]->initialCost = arcTable[i]->rental;
	for (i = 4*linkNum + 1; i < 4*linkNum+4*sourceNum; i += 2)
		arcTable[i]->initialCost = arcTable[i]->rental;
}

/** used to record flow paths after minimum cost maximum flow algorithm finished, this function should be called only once. */
void Graph::recordFlowPath()
{
	int source = superSourceID;
	int terminal = superSinkID;
	std::vector<int> path;
	path.reserve(16);
	while ( _DFS(source, terminal, path, INF) )
		;
}

/** called by recordFlowPath(). */
bool Graph::_DFS(int s, int t, std::vector<int>& path, int tmpFlow)
{
	if (s == t)
	{
		path.push_back(tmpFlow);
		fluxPathList.push_back(path);
		path.clear();
		minFlow = tmpFlow;
		return true;
	}
	struct Arc *parc = NodeList[s].firstArc;
	while (parc != NULL)
	{
		if (parc->bandwidth - parc->initialCap > 0)
		{
			if (parc->dstID != t)
				path.push_back(parc->dstID);
			if ( _DFS(parc->dstID, t, path, std::min(tmpFlow, parc->bandwidth - parc->initialCap)) )
			{
				parc->initialCap += minFlow;
				return true;
			}
		}
		parc = parc->nextArc;
	}
	return false;
}

/** add the super source node, this function should be called every time right before algorithm starts. */
void Graph::addSuperSource(std::list<int>& servers)
{
	struct Arc *parc = NULL, *rarc = NULL;

	sourceNum = static_cast<uint32_t>(servers.size());

	/*
	 * 1. range of links that between network nodes is [0, 4*linkNum-1], and all integer indices are filled;
	 * 2. range of links that between access network node and consumption node is
	 *    [4*linkNum, 4*linkNum+4*consumptionNodeNum-2], and all even indices are filled;
	 * 3. range of links that between consumption nodes and super sink node is
	 *    [4*linkNum+4*consumptionNodeNum, 4*linkNum+8*consumptionNodeNum-2], and all even indices are filled;
	 * 4. in order to use unfilled indices effectively, let range of links that between
	 *    super source node and server deployment nodes be [4*linkNum+1, 4*linkNum+4*sourceNum-1].
	 */
	arcIDIncrement = 4*linkNum + 1;

	for (std::list<int>::iterator iter = servers.begin(); iter != servers.end(); ++iter)
	{
		parc = new Arc;
		parc->srcID = superSourceID;
		parc->dstID = *iter;         // network node which deploys a CDN server
		parc->bandwidth = INF;       // throughput of the CDN server is infinite
		parc->rental = 0;            // access link need no rental fee
		parc->initialCap = INF;
		parc->initialCost = 0;
		parc->nextArc = NULL;
		rarc = new Arc;
		rarc->srcID = parc->dstID;
		rarc->dstID = parc->srcID;
		rarc->rental = 0;
		parc->reverseArc = rarc;
		rarc->reverseArc = parc;

		if (NodeList[superSourceID].firstArc == NULL)
			NodeList[superSourceID].firstArc = parc;
		else
			NodeList[superSourceID].lastArc->nextArc = parc; // append to the last
		NodeList[superSourceID].lastArc = parc; // update lastArc too

		arcTable[parc->arcID] = parc;
		arcTable[rarc->arcID] = rarc;
	}
}

/** delete the super source node, this function should be called every time right after algorithm finished. */
void Graph::delSuperSource()
{
	struct Arc *freearc1 = NodeList[superSourceID].firstArc, *freearc2 = NULL;

	if (freearc1 != NULL)
		freearc2 = freearc1->nextArc;
	else
		warning_log("have you add super source before? Or are you trying to delete super source twice?\n");

	uint32_t rank = 0;

	while (freearc1 != NULL && freearc2 != NULL)
	{
		if (rank % 2 == 0)
		{
			delete freearc1->reverseArc;
			delete freearc1;
			freearc1 = freearc2->nextArc;
		}
		else
		{
			delete freearc2->reverseArc;
			delete freearc2;
			freearc2 = freearc1->nextArc;
		}
		++rank;
	}
	if (freearc1 != NULL)
	{
		delete freearc1->reverseArc;
		delete freearc1;
	}
	if (freearc2 != NULL)
	{
		delete freearc2->reverseArc;
		delete freearc2;
	}
	
	NodeList[superSourceID].firstArc = NULL;
	NodeList[superSourceID].lastArc = NULL;

	for (uint32_t i = 4*linkNum + 1; i < 4*linkNum+4*sourceNum; i += 2)
		arcTable[i] = NULL;
}

/** display arcs from super source to deployment node, only useful when debugging. */
void Graph::displaySuperSource()
{
	if (log_level < DEBUG_LEVEL)
		return;

	printf("Super source node's adjacency list displays: \n");

	struct Arc *darc;
	printf("node: [%d]", NodeList[superSourceID].ID);
	darc = NodeList[superSourceID].firstArc;
	while (darc != NULL)
	{
		printf(" -> %d(%d,%d)", darc->dstID, darc->bandwidth, darc->rental);
		darc = darc->nextArc;
	}
	printf("\n");
}

/** add the super sink node, this function is called by preMinimumCostMaximumFlow(). */
void Graph::_addSuperSink()
{
	struct Arc *parc = NULL, *rarc = NULL;
	
	arcIDIncrement = 4*linkNum + 4*consumptionNodeNum;

	for (uint32_t i = 0; i < consumptionNodeNum; ++i)
	{
		uint32_t curNode = networkNodeNum + i; // consumption node index

		parc = new Arc;
		parc->srcID = curNode;
		parc->dstID = superSinkID;         // super sink node
		parc->bandwidth = demandTable[i];  // demand bandwidth of each consumption node
		parc->rental = 0;                  // access link need no rental fee
		parc->initialCap = demandTable[i];
		parc->initialCost = 0;
		parc->nextArc = NULL;
		rarc = new Arc;
		rarc->srcID = parc->dstID;
		rarc->dstID = parc->srcID;
		rarc->rental = 0;
		parc->reverseArc = rarc;
		rarc->reverseArc = parc;

		NodeList[curNode].firstArc = parc; // consumption node firstArc must be NULL before
		NodeList[curNode].lastArc = parc;  // update lastArc too

		arcTable[parc->arcID] = parc;
		arcTable[rarc->arcID] = rarc;
	}
}

/** delete the super sink node, this function is called by postMinimumCostMaximumFlow(). */
void Graph::_delSuperSink()
{
	uint32_t i = 0;

	for (i = 0; i < consumptionNodeNum; ++i)
	{
		struct Arc *freearc = NodeList[networkNodeNum+i].firstArc;

		if (freearc != NULL) // avoid crashing when repeat call this function
		{
			delete freearc->reverseArc;
			delete freearc;
		}

		NodeList[networkNodeNum+i].firstArc = NULL;
		NodeList[networkNodeNum+i].lastArc = NULL;
	}

	for (i = 4*linkNum+4*consumptionNodeNum; i < 4*linkNum+8*consumptionNodeNum; i += 2)
		arcTable[i] = NULL;
}

/**
 * use BFS algorithm to traverse the graph at the start node src.
 */
void Graph::BFS(int src)
{
	int i, u, v;
	for (i = 0; i < static_cast<int>(nodeNum); ++i)
	{
		color[i] = Colortype::White;
		dist[i] = INF;
		pi[i] = -1;
	}
	color[src] = Colortype::Gray;
	dist[src] = 0;
	pi[src] = -1;
	arcRecord[src] = NULL;

	std::queue<int> Q;
	Q.push(src);
	while (!Q.empty())
	{
		u = Q.front();
		Q.pop();
		struct Arc *parc = NodeList[u].firstArc;
		while (parc != NULL)
		{
			v = parc->dstID;
			if (color[v] == Colortype::White && parc->rental != INF)
			{
				color[v] = Colortype::Gray;
				dist[v] = dist[u] + 1;
				pi[v] = u;
				arcRecord[v] = parc;
				Q.push(v);
			}
			parc = parc->nextArc;
		}
		color[u] = Colortype::Black;
	}
}

/**
 * use dijkstra algorithm to calculate the minimum distance from node src to other nodes.
 */
void Graph::dijkstra(int src)
{
	int i, u, v;
	for (i = 0; i < static_cast<int>(nodeNum); ++i)
	{
		dist[i] = INF;
		pi[i] = 0;
	}
	dist[src] = 0;
	pi[src] = -1;
	arcRecord[src] = NULL;

	std::priority_queue<std::pair<int, int>, std::vector<std::pair<int, int> >, std::greater<std::pair<int, int> > > priorityQ;
	priorityQ.push(std::pair<int, int>(0, src));
	while (!priorityQ.empty())
	{
		u = priorityQ.top().second;
		priorityQ.pop();
		struct Arc *parc = NodeList[u].firstArc;
		while (parc != NULL)
		{
			v = parc->dstID;
			if (dist[v] > dist[u] + parc->rental)
			{
				dist[v] = dist[u] + parc->rental;
				pi[v] = u;
				arcRecord[v] = parc;
				priorityQ.push(std::pair<int, int>(dist[v], v));
			}
			parc = parc->nextArc;
		}
	}
}

/**
 * obtain the node path from src to dst after call BFS() or dijkstra().
 */
void Graph::recordNodePath(std::vector<int>& path, int src, int dst)
{
	std::stack<int> S;
	while (dst != src)
	{
		S.push(dst);
		dst = pi[dst];
	}
	S.push(dst);
	while (!S.empty())
	{
		path.push_back(S.top());
		S.pop();
	}
}

/**
 * obtain the arc path from src to dst after call BFS() or dijkstra().
 */
long Graph::recordArcPath(std::vector<struct Arc*>& path, int src, int dst)
{
	std::stack<int> S;
	long costSum = 0;
	while (dst != src)
	{
		int tmp = dst;
		dst = arcRecord[dst]->arcID;
		S.push(dst);
		dst = pi[tmp];
	}
	while (!S.empty())
	{
		path.push_back(arcTable[S.top()]);
		costSum += arcTable[S.top()]->rental;
		S.pop();
	}
	return costSum;
}

/** verification of the final CDN server deployment scheme. */
bool verifyScheme(Graph& graph, std::list<std::vector<int> >& fluxPathList, long& totalRentalFee)
{
	if (fluxPathList.size() > 50000)
	{
		error_log("The number of flux paths has overstep the limits!\n");
		return false;
	}

	size_t i = 0;
	totalRentalFee = 0;
	std::vector<int> demandSatified(consumptionNodeNum, 0);
	std::vector<int> arcLeftBandwidth(arcTable.size(), 0);
	for (i = 0; i < arcTable.size(); ++i)
		if (arcTable[i] != NULL)
			arcLeftBandwidth[i] = arcTable[i]->bandwidth;

	int fluxPathIdx = 1;
	for (std::list<std::vector<int> >::iterator iter = fluxPathList.begin(); iter != fluxPathList.end(); ++iter, ++fluxPathIdx)
	{
		debug_log("verify flux path %d...\n", fluxPathIdx);
		
		std::vector<int>& fluxPath = *iter; // alias
		if (fluxPath.size() < 3)
		{
			error_log("The number of nodes in a flux path less than 3, it must be an invalid flux path!\n");
			return false;
		}
		if (fluxPath.size() > 1000)
		{
			error_log("The number of nodes in a flux path has overstep the limits!\n");
			return false;
		}

		struct Arc *parc = graph.NodeList[fluxPath[0]].firstArc;
		size_t accessNodeIdx = fluxPath.size() - 3;
		int consumptionNodeID = fluxPath[accessNodeIdx+1];
		int consumptionBandwidth = fluxPath.back();
		for (i = 1; i <= accessNodeIdx; ++i)
		{
			while (parc != NULL)
			{
				if (parc->dstID == fluxPath[i])
				{
					totalRentalFee += parc->rental*consumptionBandwidth;
					arcLeftBandwidth[parc->arcID] -= consumptionBandwidth;
					if (arcLeftBandwidth[parc->arcID] < 0)
					{
						error_log("The left bandwidth of arc[%d] which is %d -> %d belows zero!\n", parc->arcID, parc->srcID, parc->dstID);
						return false;
					}
					parc = graph.NodeList[parc->dstID].firstArc;
					break;
				}
				parc = parc->nextArc;
			}
			if (parc == NULL)
			{
				error_log("The flux path has an arc that is not in the graph!\n");
				return false;
			}
		}

		if (fluxPath[accessNodeIdx] != accessTable[consumptionNodeID])
		{
			error_log("The access network node[%d] in the flux path is not the same with that required[%d]!\n", fluxPath[accessNodeIdx], accessTable[consumptionNodeID]);
			return false;
		}
		demandSatified[consumptionNodeID] += consumptionBandwidth;
	}
	
	for (uint32_t k = 0; k < consumptionNodeNum; ++k)
	{
		if (demandSatified[k] < demandTable[k])
		{
			error_log("The flux scheme cannot satify the bandwidth requirement of consumption node[%u]!\n", k);
			return false;
		}
	}
	
	return true;
}
