/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: Graph.h                                                                    *
 * @author: Xu Le, He Jianfan                                                        *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#ifndef __GRAPH_H__
#define __GRAPH_H__

#include "Utils.h"

#define INF    0x7fffffff

/** arc structure of graph. */
struct Arc
{
	Arc();
	
	int arcID;       ///< ID of this arc.
	int srcID;       ///< tail of this arc.
	int dstID;       ///< head of this arc.
	int bandwidth;   ///< max capacity of this arc.
	int rental;      ///< cost weight of this arc.
	int initialCap;  ///< initial capacity of this arc.
	int initialCost; ///< initial cost of this arc.
	Arc *nextArc;    ///< next out degree arc of the node this arc derived.
	Arc *reverseArc; ///< reverse arc used by minimum cost maximum flow algorithm.
};

/** node structure of graph. */
struct Node
{
	Node() : ID(0), firstArc(NULL), lastArc(NULL) {}
	
	int ID;        ///< ID of this node.
	Arc *firstArc; ///< first out degree arc of this node.
	Arc *lastArc;  ///< last out degree arc of this node.
};

/** adjacency list structure representing the graph. */
class Graph
{
public:
	explicit Graph(uint32_t _node_num);
	~Graph();

	/** @name regular methods of Graph. */
	///@{
	void construct(std::list<struct LinkTuple>& linkTuples);
	void undirectify(std::list<struct LinkTuple>& linkTuples);
	void destruct();
	void display();
	///@}

	/** @name API function members used by minimum cost maximum flow algorithm. */
	///@{
	void preferZKW(bool _useZKW);
	void addSuperSource(std::list<int>& servers);
	void delSuperSource();
	void displaySuperSource();
	void preMinimumCostMaximumFlow();
	long minimumCostMaximumFlow(std::list<int>& servers);
	void postMinimumCostMaximumFlow();
	void revertStatusToInitialization();
	void recordFlowPath();
	///@}

	/** @name API function members used by TopologyScanner. */
	///@{
	void BFS(int src);
	void dijkstra(int src);
	void recordNodePath(std::vector<int>& path, int src, int dst);
	long recordArcPath(std::vector<struct Arc*>& path, int src, int dst);
	///@}

public:
	/** struct used by exchange operation in Greedy algorithm. */
	class FluxFee
	{
	public:
		int from;
		int to;
		int fee;

		bool operator<(FluxFee rhs) const { return fee < rhs.fee; }
	};

	std::priority_queue<FluxFee> fluxFeeQueue; ///< used by Greedy.
	void setExchangeOperation(bool _exchange_operation);

private:
	enum Colortype
	{
		White = 0, ///< denote this node has not been visited.
		Gray,      ///< denote this node has been visited but its all out degrees hasn't been visited.
		Black      ///< denote this node's all out degrees has been visited.
	};

	Graph(const Graph&);
	Graph& operator=(const Graph&);

	/** @name internal function members used by minimum cost maximum flow algorithm. */
	///@{
	long _SPFA();
	long _ZKW();
	int _augment(int u, int f, int from, long& totalCost);
	bool _modifyLabel();
	void _addSuperSink();
	void _delSuperSink();
	void _revertCapacityToInitialization();
	void _revertCostToInitialization();
	bool _DFS(int s, int t, std::vector<int>& path, int tmpFlow);
	///@}

public:
	uint32_t nodeNum;       ///< maintaining the node number of the graph.
	struct Node *NodeList;  ///< storing all the nodes in the graph.
	Colortype *color;       ///< recording a node's visited status when BFS or DFS traverse is called.
	int *dist;              ///< recording the time when a node is discovered.
	int *f;                 ///< recording the time when a node's all out degrees has been visited.
	int *pi;                ///< maintaining relationship of node's predecessor when BFS, DFS or dijkstra traverse is called.
	struct Arc **arcRecord; ///< maintaining relationship of arc's predecessor when to call recordArcPath method after BFS or dijkstra traverse is called.

	/** @name data members used by minimum cost maximum flow algorithm. */
	///@{
	bool useZKW;            ///< true: use zkw algorithm as minimum cost flow algorithm, false: use spfa algorithm as minimum cost flow algorithm.
	bool exchangeOperation; ///< indicate whether the exchange operation in greedy algorithm has begun.
	bool *visited;          ///< used by zkw.
	long cost;              ///< used by zkw.
	FluxFee *fluxFeeArray;  ///< used by Greedy.
	FluxFee **feeMatrix;    ///< used by Greedy.
	int sumFlux;            ///< sum flux in graph, which is equals to sum up demand flux of all consumption node.
	int minFlow;            ///< used by _DFS function as a global variable.
	int superSourceID;      ///< ID of super source.
	int superSinkID;        ///< ID of super sink.
	uint32_t sourceNum;     ///< stores the number of chosen nodes where will be deployed with a server for reverting capacity of arcs that from super source to these chosen nodes.
	std::list<std::vector<int> > fluxPathList; ///< stores all flux paths from super source to super sink, which is certainly the result of this program.
	///@}
};

bool verifyScheme(Graph& graph, std::list<std::vector<int> >& fluxPathList, long& totalRentalFee);

#endif /* __GRAPH_H__ */
