/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: Greedy.cpp                                                                 *
 * @author: Xu Le, He Jianfan                                                        *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#include "Greedy.h"

#define JOIN_FOP_FOM_AS_ATOM_OPERATION    0

#define REMOVE_ACCESS_NODE    for (itS = servers.begin(); itS != servers.end(); ++itS) { if (*itS == cspRank[curIdx]) { servers.erase(itS); break; } }

extern int log_level;

extern uint32_t networkNodeNum;
extern uint32_t linkNum;
extern uint32_t consumptionNodeNum;
extern uint32_t serverDeploymentFee;

extern std::vector<int> accessTable;
extern std::vector<int> demandTable;
extern std::list<int> bestServers;

Greedy::Greedy(Graph& _graph, std::list<std::vector<int> >& _serverGroups) : graph(_graph), progress(Progress::ZOM), curIdx(0), curNBIdx(0), hasExecutedZOP(false), rightAfterExchange(true), inExchangeProgress(false)
{
	std::vector<int> &firstGroup = _serverGroups.front(); // alias
	if (firstGroup.size() != 1 || firstGroup[0] != -1)
		for (size_t i = 0; i < firstGroup.size(); ++i)
			definiteServers.insert(firstGroup[i]);
	_serverGroups.pop_front();

	for (uint32_t k = 0; k < consumptionNodeNum; ++k)
		accessSet.insert(accessTable[k]);

	deployed = new bool[networkNodeNum];
	memset(deployed, false, networkNodeNum*sizeof(bool));
	cspCN = consumptionNodeNum - static_cast<uint32_t>(definiteServers.size());
	totalCN = cspCN;
	cspRank = new int[cspCN];
	cspRDFR = new double[cspCN];
}

Greedy::~Greedy()
{
	delete []deployed;
	delete []cspRank;
	delete []cspRDFR;
}

/** API function of CDN server list initialization required by deploy.cpp. */
void Greedy::initialize(std::list<int>& servers)
{
	printf("====================    Enter greedy category    ====================\n");
	info_log("====================    Progress::ZOM(1) starts    ====================\n");

	for (uint32_t i = 1; i < cspCN; ++i) // the least rank access node has been removed
	{
		servers.push_back(cspRank[i]);
		deployed[cspRank[i]] = true;
	}

	for (std::set<int>::iterator itDef = definiteServers.begin(); itDef != definiteServers.end(); ++itDef)
	{
		servers.push_back(*itDef);
		deployed[*itDef] = true;
	}

	lastChangedNode = cspRank[curIdx];
	lastResultFee = consumptionNodeNum * serverDeploymentFee; // initial total fee only include server deployment fee
}

/**
 * @brief API function of CDN server list optimization required by deploy.cpp.
 * 
 * 1. initial state is Progress::ZOM and bool variable hasExecutedZOP == false, thus starts at line 127, when it is time to exit this state, then go to line 146,150;
 * 2. when state switched to Progress::ZOP, starts at line 172, when it is time to exit this state, set bool variable hasExecutedZOP = true, then go to line 193,205;
 * 3. when state switched back to Progress::ZOM, starts at line 97, when it is time to exit this state, then go to line 120;
 * 4. when state switched to Progress::EX, starts at line 412, when an exchange action was actually done, set bool variable inExchangeProgress = true and go to line 473, next loop will switch to sub Progress::ZOM;
 *      4.1. the same as step 3. but when it is time to exit loop, then go to line 115, next loop will go to line 146,154;
 *      4.2. the same as step 2. but when it is time to exit loop, then go to line 193,197, next loop return back to Progress::EX and then starts at line 423;
 *    when tryExchangeTimes has exceed the default set or priority_queue graph.fluxFeeQueue is empty, exit Progress::EX at line 456 and switch to state Progress::FOP the next loop;
 * 5. when state switch to Progress::FOP, starts at line 224, here will call function _selectRestNode() at line 266, the next loop will starts at line 306,
 *    when it is time to exit this state, go to line 323 and the switch state to Progress::FOM, the next loop will starts at line 335,
 *    note that this function will return false at line 346, thus the bool variable continueLoop will be assigned false and quit the loop in deploy.cpp.
 */
bool Greedy::optimize(std::list<int>& servers, long justRentalFee)
{
	long deploymentFee = static_cast<long>(servers.size()) * serverDeploymentFee;
	uint32_t i = 0;

	switch (progress)
	{
	case Progress::ZOM:
	{
		if (hasExecutedZOP) // usually will enter this part
		{
			if (lastResultFee <= justRentalFee + deploymentFee || justRentalFee == -1) // restore the most recently removed node
			{
				servers.push_front(lastChangedNode);
				deployed[lastChangedNode] = true;
			}
			else
				lastResultFee = justRentalFee + deploymentFee;

			while (!deployed[cspRank[curIdx]] && curIdx < static_cast<int>(cspCN-1))
				++curIdx;
			if (curIdx < static_cast<int>(cspCN-1)) // quit above while loop due to the first condition
			{
				REMOVE_ACCESS_NODE
				deployed[cspRank[curIdx]] = false;
				lastChangedNode = cspRank[curIdx++];
			}
			else if (inExchangeProgress)
			{
				curIdx = -1; // prepare to start ZOP progress
				hasExecutedZOP = false;
			}
			else
			{
				curIdx = 0;
				graph.setExchangeOperation(true); // needed by exchange progress
				progress = Progress::EX;
			}
		}
		else if (curIdx != -1) // only the initial state ZOM will enter this part
		{
			if (lastResultFee <= justRentalFee + deploymentFee || justRentalFee == -1) // restore the most recently removed node
			{
				servers.push_front(lastChangedNode);
				deployed[lastChangedNode] = true;
			}
			else
				lastResultFee = justRentalFee + deploymentFee;

			if (++curIdx >= static_cast<int>(cspCN-1))
				curIdx = -1; // prepare to start ZOP progress
			else
			{
				REMOVE_ACCESS_NODE
				deployed[cspRank[curIdx]] = false;
				lastChangedNode = cspRank[curIdx];
			}
		}
		else // curIdx == -1, actually here is a transitional state when the next loop will switch to Progress::ZOP
		{
			lastResultFee = justRentalFee + deploymentFee;

			if (!inExchangeProgress)
			{
				info_log("Total fee after Progress::ZOM(1) is %ld\n\n", lastResultFee);
				info_log("====================    Progress::ZOP starts    ====================\n");
			}
			else
				info_log("Total fee after sub Progress::ZOM is %ld\n", lastResultFee);

			curIdx = 0; // return to 0, loop restarts again
			while (deployed[cspRank[curIdx]])
				++curIdx;

			servers.push_front(cspRank[curIdx]);
			deployed[cspRank[curIdx]] = true;
			lastChangedNode = cspRank[curIdx++];

			progress = Progress::ZOP;
		}
		break;
	}
	case Progress::ZOP:
	{
		if (!hasExecutedZOP) // each time when state has just switched to zero order plus, hasExecutedZOP is always false, see line 116
		{
			if (lastResultFee <= justRentalFee + deploymentFee || justRentalFee == -1) // remove the most recently added node
			{
				servers.pop_front();
				deployed[lastChangedNode] = false;
			}
			else
				lastResultFee = justRentalFee + deploymentFee;

			while (deployed[cspRank[curIdx]] && curIdx < static_cast<int>(cspCN-1))
				++curIdx;
			if (curIdx < static_cast<int>(cspCN-1)) // quit above while loop due to the first condition
			{
				servers.push_front(cspRank[curIdx]);
				deployed[cspRank[curIdx]] = true;
				lastChangedNode = cspRank[curIdx++];
			}
			else
				hasExecutedZOP = true;
		}
		else
		{
			lastResultFee = justRentalFee + deploymentFee;

			if (inExchangeProgress) // return to exchange progress
			{
				info_log("Total fee after sub Progress::ZOP is %ld\n", lastResultFee);

				curIdx = 1; // assign any number that larger than 0
				graph.setExchangeOperation(true); // update the whole fluxFeeQueue for next exchange preparation
				progress = Progress::EX;
			}
			else // go to zero order progress
			{
				curIdx = 0; // return to 0, loop restarts again
				while (!deployed[cspRank[curIdx]])
					++curIdx;

				REMOVE_ACCESS_NODE
				deployed[cspRank[curIdx]] = false;
				lastChangedNode = cspRank[curIdx++];

				progress = Progress::ZOM;
				info_log("Total fee after Progress::ZOP is %ld\n\n", lastResultFee);
				info_log("====================    Progress::ZOM(2) starts    ====================\n");
			}
		}
		break;
	}
	case Progress::FOP:
	{
		if (curIdx == 0 && curNBIdx == 0)
		{
			lastResultFee = justRentalFee + deploymentFee;

			// return false;
			info_log("Total fee after Progress::EX is %ld\n\n", lastResultFee);
			info_log("====================    Progress::FOP starts    ====================\n");

			// resort consumption node
			uint32_t newIdx = 0, abandonIdx = 0;
			int *abandonRank = new int[totalCN + definiteServers.size() - servers.size()];
			double *abandonRDFR = new double[totalCN + definiteServers.size() - servers.size()];
			for (i = 0; i < totalCN; ++i)
			{
				if (deployed[cspRank[i]])
				{
					cspRank[newIdx] = cspRank[i];
					cspRDFR[newIdx] = cspRDFR[i];
					++newIdx;
				}
				else
				{
					abandonRank[abandonIdx] = cspRank[i];
					abandonRDFR[abandonIdx] = cspRDFR[i];
					++abandonIdx;
				}
			}
			if (totalCN + definiteServers.size() - servers.size() != abandonIdx)
			{
				error_log("assert failed.\n");
				exit(EXIT_FAILURE);
			}
			cspCN = newIdx; // update cspCN
			for (i = cspCN; i < totalCN; ++i)
			{
				cspRank[i] = abandonRank[i-cspCN];
				cspRDFR[i] = abandonRDFR[i-cspCN];
			}
			delete []abandonRank;
			delete []abandonRDFR;

			for (i = 0; i < totalCN; ++i)
				debug_log("[%u] -> cspRank: %d, cspRDFR: %f\n", i, cspRank[i], cspRDFR[i]);

			neighborCN = _selectRestNode(false);

			if (neighborCN == 0)
				return false;

			if (log_level >= INFO_LEVEL)
			{
				printf("INFO |");
				for (i = 0; i < neighborCN; ++i)
					printf(" %d", neighborRank[i]);
				printf("\n");
			}

			servers.push_front(neighborRank[0]); // curIdx == curNBIdx == 0
			deployed[neighborRank[0]] = true;
#if JOIN_FOP_FOM_AS_ATOM_OPERATION
			progress = Progress::FOM; // note that curIdx still equals to 0
#else
			lastChangedNode = neighborRank[curIdx++];
#endif
		}
		else
		{
#if JOIN_FOP_FOM_AS_ATOM_OPERATION
			curIdx = 0; // test starts from least access node
			info_log("curNBIdx: %d\n", curNBIdx);
			if (curNBIdx == static_cast<int>(neighborCN)) // all selected neighbor nodes have been tested
			{
				delete []neighborRank;

				printf("====================    Exit greedy category    ====================\n\n");
				return false; // quit FOP-FOM atom operation
			}
			else
			{
				servers.push_front(neighborRank[curNBIdx]);
				deployed[neighborRank[curNBIdx]] = true;
				progress = Progress::FOM; // starts FOM operation
			}
#else
			if (lastResultFee <= justRentalFee + deploymentFee || justRentalFee == -1) // restore the most recently added neighbor node
			{
				servers.pop_front();
				deployed[lastChangedNode] = false;
			}
			else
			{
				lastResultFee = justRentalFee + deploymentFee;
				info_log("a neighbor[%d] pass the test\n", lastChangedNode);
			}

			servers.push_front(neighborRank[curIdx]);
			deployed[neighborRank[curIdx]] = true;
			lastChangedNode = neighborRank[curIdx];

			if (++curIdx == static_cast<int>(neighborCN))
			{
				curIdx = -1; // return to -1, loop restarts again
				progress = Progress::FOM;
			}
#endif
		}
		break;
	}
	case Progress::FOM:
	{
		if (curIdx == -1)
		{
#if !JOIN_FOP_FOM_AS_ATOM_OPERATION
			if (lastResultFee <= justRentalFee + deploymentFee || justRentalFee == -1) // restore the most recently added neighbor node
			{
				servers.pop_front();
				deployed[lastChangedNode] = false;
			}
			else
				lastResultFee = justRentalFee + deploymentFee;

			delete []neighborRank;

			printf("====================    Exit greedy category    ====================\n\n");
			return false;
#endif
		}
		else if (curIdx == 0)
		{
#if JOIN_FOP_FOM_AS_ATOM_OPERATION
			if (lastResultFee < justRentalFee + deploymentFee || justRentalFee == -1) // remove the least access node
			{
				while (!deployed[cspRank[curIdx]]) // ensure the current access node to remove soon is deployed with a server now
					++curIdx;
				REMOVE_ACCESS_NODE
				deployed[cspRank[curIdx]] = false;
				lastChangedNode = cspRank[curIdx];
				while (!deployed[cspRank[curIdx]]) // here will enter loop at least once, ensure the next access node to remove is deployed with a server now
					++curIdx;
			}
			else
			{
				lastResultFee = justRentalFee + deploymentFee;
				info_log("a neighbor[%d] pass the test\n", neighborRank[curNBIdx]);
				++curNBIdx;
				progress = Progress::FOP; // return to progress FOP, then test the next neighbor node
			}
#endif
		}
		else
		{
#if JOIN_FOP_FOM_AS_ATOM_OPERATION
			if (lastResultFee < justRentalFee + deploymentFee || justRentalFee == -1) // remove the least access node
			{
				REMOVE_ACCESS_NODE
				deployed[cspRank[curIdx]] = false;
				
				servers.push_front(lastChangedNode);
				deployed[lastChangedNode] = true;
				lastChangedNode = cspRank[curIdx];
				while (!deployed[cspRank[curIdx]] && curIdx < static_cast<int>(cspCN)) // here will enter loop at least once, ensure the next access node to remove is deployed with a server now
					++curIdx;
				if (curIdx == static_cast<int>(cspCN))
				{
					// remove current neighbor node added before, because this neighbor node cannot decrease the total fee and failed to pass the test
					for (itS = servers.begin(); itS != servers.end(); ++itS) { if (*itS == neighborRank[curNBIdx]) { servers.erase(itS); break; } }

					servers.push_front(lastChangedNode); // note lastChangedNode has been assigned at 8 lines up here, it is not neccessary to test the last access node
					deployed[lastChangedNode] = true;

					++curNBIdx;
					progress = Progress::FOP; // return to progress FOP, then test the next neighbor node
				}
			}
			else
			{
				lastResultFee = justRentalFee + deploymentFee;
				info_log("a neighbor[%d] pass the test\n", neighborRank[curNBIdx]);
				++curNBIdx;
				progress = Progress::FOP; // return to progress FOP, then test the next neighbor node
			}
#endif
		}
		break;
	}
	case Progress::EX:
	{
		Graph::FluxFee fluxFee;
		if (curIdx == 0 || rightAfterExchange)
		{
			if (curIdx == 0)
			{
				lastResultFee = justRentalFee + deploymentFee;

				info_log("Total fee after Progress::ZOM(2) is %ld\n\n", lastResultFee);
				info_log("====================    Progress::EX starts    ====================\n");

				inExchangeProgress = true;
				tryExchangeTimes = 0;
			}

			fluxFee = graph.fluxFeeQueue.top();
			graph.fluxFeeQueue.pop();

			debug_log("from: %d, to: %d, fee: %d\n", fluxFee.from, fluxFee.to, fluxFee.fee);
			// exchange from and to, that is, remove 'from' node as well as add 'to' node
			for (itS = servers.begin(); itS != servers.end(); ++itS) { if (*itS == fluxFee.from) { servers.erase(itS); break; } }
			servers.push_front(fluxFee.to);
			deployed[fluxFee.from] = false;
			deployed[fluxFee.to] = true;
			lastChangedNode = fluxFee.from;
			graph.setExchangeOperation(false); // avoid changing fluxFeeQueue
			rightAfterExchange = false;

			++curIdx;
		}
		else
		{
			if (lastResultFee <= justRentalFee + deploymentFee || justRentalFee == -1) // undo previous exchange operation
			{
				deployed[servers.front()] = false;
				servers.pop_front(); // remove the 'to' node added in previous exchange operation
				servers.push_front(lastChangedNode); // restore the 'from' node removed in previous exchange operation
				deployed[lastChangedNode] = true;

				if (graph.fluxFeeQueue.empty() || tryExchangeTimes > 1000) // exit EX progress
				{
					info_log("tryExchangeTimes: %d\n", tryExchangeTimes);

					curIdx = 0;
					while (!graph.fluxFeeQueue.empty())
						graph.fluxFeeQueue.pop();
					graph.setExchangeOperation(false); // no need to change fluxFeeQueue anymore
					inExchangeProgress = false; // after now, zero order operation is not a sub operation of exchange operation anymore
					progress = Progress::FOP;
					return true;
				}


				fluxFee = graph.fluxFeeQueue.top();
				graph.fluxFeeQueue.pop();

				debug_log("from: %d, to: %d, fee: %d\n", fluxFee.from, fluxFee.to, fluxFee.fee);
				for (itS = servers.begin(); itS != servers.end(); ++itS) { if (*itS == fluxFee.from) { servers.erase(itS); break; } } // exchange from and to
				servers.push_front(fluxFee.to);
				deployed[fluxFee.from] = false;
				deployed[fluxFee.to] = true;
				lastChangedNode = fluxFee.from;
			}
			else
			{
				lastResultFee = justRentalFee + deploymentFee;
				info_log("node [%d] and [%d] exchanged, now total fee is %ld\n", lastChangedNode, servers.front(), lastResultFee);

				rightAfterExchange = true;

				// prepare to begin sub zero order minus progress
				curIdx = 0; // return to 0, loop restarts again
				while (!deployed[cspRank[curIdx]])
					++curIdx;

				REMOVE_ACCESS_NODE
				deployed[cspRank[curIdx]] = false;
				lastChangedNode = cspRank[curIdx++];

				progress = Progress::ZOM;
			}

			++tryExchangeTimes;
		}
		break;
	}
	default:
		error_log("Unknown progress!\n");
	}

	return true;
}

void Greedy::obtainOutLinkCap(int *_out_link_cap)
{
	outLinkCap = _out_link_cap;
}

void Greedy::obtainRentalDeployFeeRatio(double *_rental_deploy_fee_ratio)
{
	_sortCspNode(_rental_deploy_fee_ratio);
}

void Greedy::_sortCspNode(double *_rental_deploy_fee_ratio)
{
	std::vector<std::pair<double /* RDFR */, int /* cspID */> > cspSortVec;
	cspSortVec.reserve(cspCN);

	uint32_t i = 0;
	for (i = 0; i < consumptionNodeNum; ++i)
		if (definiteServers.find(accessTable[i]) == definiteServers.end())
			cspSortVec.emplace_back(demandTable[i], accessTable[i]);

	std::sort(cspSortVec.begin(), cspSortVec.end(), [](const std::pair<int, int> lhs, const std::pair<int, int> rhs) { return lhs.first < rhs.first; } );

	for (i = 0; i < cspCN; ++i)
	{
		cspRank[i] = cspSortVec[i].second;
		cspRDFR[i] = cspSortVec[i].first;
		debug_log("[%u] -> cspRank: %d, cspRDFR: %f\n", i, cspRank[i], cspRDFR[i]);
	}
}

uint32_t Greedy::_selectRestNode(bool onlyAbandon)
{
	uint32_t i = 0, NB_NUM = 0;

	std::vector<std::pair<int /* capacity */, int /* ID */> > neighbors;

	{
		std::set<int> neighborSet;
#if 0
		uint32_t beginIdx = onlyAbandon ? cspCN : 0;
		struct Arc *parc = NULL;
		for (i = beginIdx; i < totalCN; ++i)
			for (parc = graph.NodeList[cspRank[i]].firstArc; parc != NULL; parc = parc->nextArc)
				if (parc->dstID < static_cast<int>(networkNodeNum) && accessSet.find(parc->dstID) == accessSet.end()) // this neighbor is neither a consumption node nor an access node
					neighborSet.insert(parc->dstID);
#endif
		for (int j = 0; j < static_cast<int>(networkNodeNum); ++j)
			if (accessSet.find(j) == accessSet.end())
				neighborSet.insert(j);
		NB_NUM = static_cast<uint32_t>(neighborSet.size());
		neighbors.reserve(NB_NUM);
		for (std::set<int>::iterator itNS = neighborSet.begin(); itNS != neighborSet.end(); ++itNS)
			neighbors.emplace_back(0, *itNS);
	}

	info_log("NB_NUM: %u\n", NB_NUM);

	uint32_t SELECTED_NB_NUM = 0;

	for (i = 0; i < NB_NUM; ++i)
		neighbors[i].first = outLinkCap[neighbors[i].second];

	SELECTED_NB_NUM = NB_NUM; // this can be modified

#if 0
	uint32_t abandonCN = totalCN - cspCN;
	uint32_t countedCN = onlyAbandon ? abandonCN : totalCN;
	int *hopAmongNeighborsAndAbandons = new int[NB_NUM*countedCN];
	int **hopAmong = new int*[NB_NUM]; // a point array, convenient to use form hopAmong[i][j]
	for (i = 0; i < NB_NUM; ++i)
		hopAmong[i] = hopAmongNeighborsAndAbandons + i*countedCN;

	for (i = 0; i < NB_NUM; ++i)
	{
		graph.BFS(neighbors[i].second);
		for (uint32_t j = 0; j < countedCN; ++j)
		{
			hopAmong[i][j] = onlyAbandon ? graph.dist[cspRank[j+cspCN]] : graph.dist[cspRank[j]];
			if (hopAmong[i][j] == 1)
				++neighbors[i].first;
			if (log_level >= DEBUG_LEVEL)
				printf("%d ", hopAmong[i][j]);
		}
		if (log_level >= DEBUG_LEVEL)
			printf(" ->  %d\n", neighbors[i].first);
		if (neighbors[i].first > 1)
			++SELECTED_NB_NUM;
	}
	delete []hopAmong;
	delete []hopAmongNeighborsAndAbandons;

	info_log("SELECTED_NB_NUM: %u\n", SELECTED_NB_NUM);

	if (SELECTED_NB_NUM == 0)
	{
		warning_log("Here is no neighbor node meet the requirement.\n");
		neighborRank = NULL; // C++ standrad promise it is safe to delete nullptr
		return 0;
	}
#endif

	std::sort(neighbors.begin(), neighbors.end(), std::greater<std::pair<int, int> >());

	neighborRank = new int[SELECTED_NB_NUM];
	for (i = 0; i < SELECTED_NB_NUM; ++i)
		neighborRank[i] = neighbors[i].second;

	return SELECTED_NB_NUM;
}
