/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: TopologyScanner.cpp                                                        *
 * @author: Xu Le                                                                    *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#include "TopologyScanner.h"

int maxUnitRentalFee = 0;

extern int log_level;

extern uint32_t networkNodeNum;
extern uint32_t linkNum;
extern uint32_t consumptionNodeNum;
extern uint32_t serverDeploymentFee;

extern std::vector<int> accessTable;
extern std::vector<int> demandTable;

TopologyScanner::TopologyScanner(Graph& _graph) : graph(_graph), arcDensity(ArcDensity::Unknown), cspDensity(CspDensity::Unknown), lowerBoundNum(0), multiplicativeFactor(1.0)
{
	mNetworkNodeNum = static_cast<int>(networkNodeNum);
	mServerDeploymentFee = static_cast<int>(serverDeploymentFee);
	upBoundNum = static_cast<int>(consumptionNodeNum);
	outLinkNum = new int[networkNodeNum];
	outLinkCap = new int[networkNodeNum];
	outLinkFee = new double[networkNodeNum];
	capRank = new double[networkNodeNum];
	feeRank = new double[networkNodeNum];
	scoreTable = new int[networkNodeNum];
	rentalDeployFeeRatio = new double[consumptionNodeNum];
	std::fill(outLinkNum, outLinkNum + networkNodeNum, 0);
	std::fill(outLinkCap, outLinkCap + networkNodeNum, 0);
	std::fill(outLinkFee, outLinkFee + networkNodeNum, 0.0);
}

TopologyScanner::~TopologyScanner()
{
	delete []outLinkNum;
	delete []outLinkCap;
	delete []outLinkFee;
	delete []capRank;
	delete []feeRank;
	delete []scoreTable;
	delete []rentalDeployFeeRatio;
}

/** scan the network topology of graph. */
void TopologyScanner::scan()
{
	printf("\n====================    Enter topology scanner category    ====================\n");

	_defineDensity();

	uint32_t i = 0;

	for (i = 0; i < networkNodeNum; ++i)
	{
		struct Arc *parc = graph.NodeList[i].firstArc;
		while (parc != NULL)
		{
			if (parc->dstID < mNetworkNodeNum)
			{
				++outLinkNum[i];
				outLinkCap[i] += parc->bandwidth;
				outLinkFee[i] += parc->bandwidth*parc->rental;
			}
			parc = parc->nextArc;
		}
	}
	std::vector<std::pair<int, int> > sortLinkCap;
	std::vector<std::pair<double, int> > sortLinkFee;

	for (i = 0; i < networkNodeNum; ++i)
	{
		outLinkFee[i] /= outLinkCap[i];
		sortLinkCap.emplace_back(outLinkCap[i], i);
		sortLinkFee.emplace_back(outLinkFee[i], i);
	}
	std::sort(sortLinkCap.begin(), sortLinkCap.end(), std::greater<std::pair<int, int> >());
	std::sort(sortLinkFee.begin(), sortLinkFee.end());
	for (i = 0; i < networkNodeNum; ++i)
	{
		capRank[sortLinkCap[i].second] = 1.0 - static_cast<double>(i)/mNetworkNodeNum;
		feeRank[sortLinkFee[i].second] = 1.0 - static_cast<double>(i)/mNetworkNodeNum;
	}
	for (i = 0; i < networkNodeNum; ++i)
	{
		debug_log("node[%u] -> outLinkCap: %d, outLinkFee: %f, capRank: %f, feeRank: %f\n", i, outLinkCap[i], outLinkFee[i], capRank[i], feeRank[i]);
	}

	for (i = 0; i < networkNodeNum; ++i)
		scoreTable[i] = 200 * capRank[i] + 200 * feeRank[i];

	_mustDeployOnAccess();

	// _calcCheapestAmongAccess();

	// _scanNeighborOfAccess();

	_selectByScore();
}

/** recommend some node groups, combinations of them may be good as the input of the next algorithm. */
void TopologyScanner::recommend(std::list<std::vector<int> >& initialServerGroups)
{
	uint32_t i = 0, j = 0, definiteNum = 0;
	std::vector<int> certainGroup;
	info_log("definite servers:");
	for (size_t k = 0; k < possibleServers.size(); ++k)
	{
		if (scoreTable[possibleServers[k]] != INF)
			break;
		++definiteNum;
		certainGroup.push_back(possibleServers[k]);
		printf(" %d", possibleServers[k]);
	}
	for (i = definiteNum; i < static_cast<uint32_t>(possibleServers.size()); ++i) // shift to overlaying definite servers, pop_back() operation is not necessary
		possibleServers[i-definiteNum] = possibleServers[i];

	printf("\n");
	info_log("possible servers:");
	if (log_level >= INFO_LEVEL) // print possible servers
	{
		for (i = 0; i < static_cast<uint32_t>(possibleServers.size())-definiteNum; ++i)
			printf(" %d", possibleServers[i]);
		printf("\n");
	}
	if (definiteNum == 0)
		certainGroup.push_back(-1); // indicates there is no network nodes have to deploy a server
	initialServerGroups.push_back(certainGroup);
	certainGroup.clear();

	// divide possible servers into groups in order to use different combination of them,
	// obviously, different combination lead to different link rental fee
	uint32_t CN = static_cast<uint32_t>(possibleServers.size()) - definiteNum; // alias of short name "CandidateNumber"

	int *hopAmongCandidates = new int[CN*CN];
	int **hopAmong = new int*[CN]; // a point array, convenient to use form hopAmong[i][j]
	for (i = 0; i < CN; ++i)
		hopAmong[i] = hopAmongCandidates + i*CN;

	int *minimumHop = new int[CN];
	double *averageHop = new double[CN];
	std::unordered_set<int> accessNodeSet;
	for (i = 0; i < consumptionNodeNum; ++i)
		accessNodeSet.insert(accessTable[i]);
	std::vector<std::pair<int /* group */, bool /* determined */> > partion;
	partion.reserve(CN);
	for (i = 0; i < CN; ++i)
	{
		partion.emplace_back(0, false);
		minimumHop[i] = INF;
		averageHop[i] = 0.0;
		graph.BFS(possibleServers[i]);
		for (j = 0; j < CN; ++j)
		{
			int hops = graph.dist[possibleServers[j]]; // alias
			hopAmong[i][j] = hops;
			averageHop[i] += hops;
			if (minimumHop[i] > hops && j != i)
				minimumHop[i] = hops;
			if (log_level >= DEBUG_LEVEL)
				printf("%d ", hops);
		}
		averageHop[i] /= (CN-1);
		if (log_level >= DEBUG_LEVEL)
			printf("    ---    %d,%f\n", minimumHop[i], averageHop[i]);
	}

	int curGroup = 1;
	for (i = 0; i < CN; ++i)
	{
		if (partion[i].second == false)
		{
			partion[i].first = curGroup;
			for (j = i+1; j < CN; ++j)
			{
				if (partion[j].second == false && hopAmong[i][j] == 1 && (accessNodeSet.find(possibleServers[i]) == accessNodeSet.end() || accessNodeSet.find(possibleServers[j]) == accessNodeSet.end()))
				{
					partion[j].first = curGroup;
					partion[j].second = true;
				}
			}
			++curGroup;
		}
	}
	if (log_level >= DEBUG_LEVEL)
	{
		for (i = 0; i < CN; ++i)
			printf("%d ", partion[i].first);
		printf("\n");
	}

	for (int groupIdx = 1; groupIdx < curGroup; ++groupIdx)
	{
		for (i = 0; i < CN; ++i)
		{
			if (partion[i].first == groupIdx)
				certainGroup.push_back(possibleServers[i]);
		}
		// info_log("group %d:", groupIdx);
		// printVector(certainGroup, " ");
		initialServerGroups.push_back(certainGroup);
		certainGroup.clear();
	}

	delete []minimumHop;
	delete []averageHop;
	delete []hopAmong;
	delete []hopAmongCandidates;

	printf("====================    Exit topology scanner category    ====================\n\n");
}

const char* TopologyScanner::_arcDensityString()
{
	switch (arcDensity)
	{
	case ArcDensity::UltraSparse:
		return "ultra-sparse";
	case ArcDensity::VerySparse:
		return "very-sparse";
	case ArcDensity::Sparse:
		return "sparse";
	case ArcDensity::General:
		return "general";
	case ArcDensity::Dense:
		return "dense";
	case ArcDensity::VeryDense:
		return "very-dense";
	case ArcDensity::UltraDense:
		return "ultra-dense";
	default:
		return "unknown";
	}
}

const char* TopologyScanner::_cspDensityString()
{
	switch (cspDensity)
	{
	case CspDensity::UltraSparse:
		return "ultra-sparse";
	case CspDensity::VerySparse:
		return "very-sparse";
	case CspDensity::Sparse:
		return "sparse";
	case CspDensity::General:
		return "general";
	case CspDensity::Dense:
		return "dense";
	case CspDensity::VeryDense:
		return "very-dense";
	case CspDensity::UltraDense:
		return "ultra-dense";
	default:
		return "unknown";
	}
}

void TopologyScanner::_defineDensity()
{
	double _arc_density = static_cast<double>(2*linkNum) / mNetworkNodeNum;
	if (_arc_density < 2.0)
		arcDensity = ArcDensity::UltraSparse;
	else if (_arc_density >= 2.0 && _arc_density < 3.0)
		arcDensity = ArcDensity::VerySparse;
	else if (_arc_density >= 3.0 && _arc_density < 4.0)
		arcDensity = ArcDensity::Sparse;
	else if (_arc_density >= 4.0 && _arc_density < 6.0)
		arcDensity = ArcDensity::General;
	else if (_arc_density >= 6.0 && _arc_density < 9.0)
		arcDensity = ArcDensity::Dense;
	else if (_arc_density >= 9.0 && _arc_density < 14.0)
		arcDensity = ArcDensity::VeryDense;
	else // (_arc_density >= 14.0)
		arcDensity = ArcDensity::UltraDense;
	info_log("arc ratio: %f, its density level: %s\n", _arc_density, _arcDensityString());
		
	double _csp_density = static_cast<double>(consumptionNodeNum) / mNetworkNodeNum;
	if (_csp_density < 0.1)
		cspDensity = CspDensity::UltraSparse;
	else if (_csp_density >= 0.1 && _csp_density < 0.15)
		cspDensity = CspDensity::VerySparse;
	else if (_csp_density >= 0.15 && _csp_density < 0.2)
		cspDensity = CspDensity::Sparse;
	else if (_csp_density >= 0.2 && _csp_density < 0.3)
		cspDensity = CspDensity::General;
	else if (_csp_density >= 0.3 && _csp_density < 0.4)
		cspDensity = CspDensity::Dense;
	else if (_csp_density >= 0.4 && _csp_density < 0.5)
		cspDensity = CspDensity::VeryDense;
	else // (_csp_density >= 0.5)
		cspDensity = CspDensity::UltraDense;
	info_log("csp ratio: %f, its density level: %s\n", _csp_density, _cspDensityString());
}

void TopologyScanner::_mustDeployOnAccess()
{
	std::vector<std::pair<int /* rental */, int /* bandwidth */> > feeCapPair;
	feeCapPair.reserve(32); // 2^5 > 20

	for (uint32_t i = 0; i < consumptionNodeNum; ++i)
	{
		int accessNode = accessTable[i]; // alias
		int demandFlux = demandTable[i]; // alias
		if (outLinkCap[accessNode] < demandFlux)
		{
			scoreTable[accessNode] = INF; // large enough, indicating this network node must be deployed with a server
			continue;
		}

		struct Arc *parc = graph.NodeList[accessNode].firstArc;
		while (parc != NULL)
		{
			if (parc->dstID < mNetworkNodeNum)
				feeCapPair.emplace_back(parc->rental, parc->bandwidth);
			parc = parc->nextArc;
		}

		std::sort(feeCapPair.begin(), feeCapPair.end(), [](const std::pair<int, int>& lhs, const std::pair<int, int>& rhs) {
			if (lhs.first == rhs.first)
				return lhs.second > rhs.second;
			return lhs.first < rhs.first;
		});

		int inputFlux = 0, inputRentalFee = 0;
		size_t j = 0;
		for (; inputFlux < demandFlux && j < feeCapPair.size(); ++j)
		{
			inputFlux += feeCapPair[j].second;
			inputRentalFee += feeCapPair[j].first * feeCapPair[j].second;
		}
		inputRentalFee -= (inputFlux - demandFlux) * feeCapPair[j-1].first;
		rentalDeployFeeRatio[i] = static_cast<double>(inputRentalFee) / mServerDeploymentFee;

		debug_log("minimum input rental fee of access node [%d] is %d, rental deploy ratio %f\n", accessNode, inputRentalFee, rentalDeployFeeRatio[i]);

		if (inputRentalFee >= mServerDeploymentFee)
		{
			info_log("access node [%d] have to deploy a server.\n", accessNode);
			scoreTable[accessNode] = INF; // large enough, indicating this network node must be deployed with a server
		}

		feeCapPair.clear();
	}

	for (uint32_t k = 0; k < consumptionNodeNum; ++k)
		if (scoreTable[accessTable[k]] != INF) // avoid int type overflow
			scoreTable[accessTable[k]] += 1000 * rentalDeployFeeRatio[k];
}

void TopologyScanner::_calcCheapestAmongAccess()
{
	uint32_t i = 0, j = 0;
	std::vector<int> nodePath;
	std::vector<struct Arc*> arcPath;
	std::set<int> leftSet, rightSet;
	for (i = 0; i < consumptionNodeNum-1; ++i)
	{
		int u = accessTable[i]; // alias
		graph.dijkstra(u);
		for (j = i+1; j < consumptionNodeNum; ++j)
		{
			int v = accessTable[j]; // alias
			graph.recordNodePath(nodePath, u, v);
			graph.recordArcPath(arcPath, u, v);
			debug_log("cheapest node path [%d] -> [%d]:", u, v);
			if (log_level >= DEBUG_LEVEL)
				printVector(nodePath, " ");
			if (scoreTable[u] != INF && scoreTable[v] != INF && arcPath.size() > 1) // most cases, neither u nor v has deployed a server
			{
				size_t k = 0;
				int leftMinCap = 0, rightMinCap = 0;
				int leftFeeSum = 0, rightFeeSum = 0;
				leftSet.insert(arcPath[0]->bandwidth);
				leftFeeSum = arcPath[0]->rental;
				for (k = 1; k < arcPath.size(); ++k)
				{
					rightSet.insert(arcPath[k]->bandwidth);
					rightFeeSum += arcPath[k]->rental;
				}
				for (k = 1; k < arcPath.size(); ++k)
				{
					leftMinCap = *leftSet.begin();
					rightMinCap = *rightSet.begin();
					if (scoreTable[nodePath[k]] != INF)
					{
						double weightedUnitFee = static_cast<double>(leftMinCap*leftFeeSum + rightMinCap*rightFeeSum) / (leftMinCap + rightMinCap);
						double calcScore = 15*maxUnitRentalFee - 10*weightedUnitFee; // TODO: should be more proper
						// debug_log("[%d] -> [%d]: node [%d]'s score [%f]\n", u, v, nodePath[k], calcScore);
						scoreTable[nodePath[k]] += static_cast<int>(calcScore);
					}
					leftSet.insert(arcPath[k]->bandwidth);
					rightSet.erase(arcPath[k]->bandwidth);
					leftFeeSum += arcPath[k]->rental;
					rightFeeSum -= arcPath[k]->rental;
				}
			}
			nodePath.clear();
			arcPath.clear();
		}
		if (log_level >= DEBUG_LEVEL)
			printf("\n");
	}
	debug_log("maxUnitRentalFee: %d\n", maxUnitRentalFee);
}

void TopologyScanner::_scanNeighborOfAccess()
{
	std::vector<bool> scanned(consumptionNodeNum, false);
	std::unordered_map<int, uint32_t> accessNodeMap;
	for (uint32_t k = 0; k < consumptionNodeNum; ++k)
		accessNodeMap.emplace(accessTable[k], k);

	for (uint32_t i = 0; i < consumptionNodeNum; ++i)
	{
		int u = accessTable[i];           // alias
		int demandFluxU = demandTable[i]; // alias

		if (scoreTable[u] == INF || scanned[i] == true) // current access node is deployed with a server, skip it
			continue;

		int inputNeighborAccessFee = 0;
		struct Arc *parc = graph.NodeList[u].firstArc, *qarc = NULL;
		while (parc != NULL)
		{
			int v = parc->dstID; // alias
			int curLinkFlux = 0;
			if (accessNodeMap.find(v) != accessNodeMap.end()) // this neighbor is also an access node
			{
				if (scoreTable[v] == INF) // this neighbor v is deployed with a server, check if current access node should deploy a server too
				{
					curLinkFlux = demandFluxU >= parc->bandwidth ? parc->bandwidth : demandFluxU;
					inputNeighborAccessFee += curLinkFlux * parc->rental;
					debug_log("inputNeighborAccessFee of [%d]: %d\n", u, inputNeighborAccessFee);
					if (inputNeighborAccessFee < mServerDeploymentFee && demandFluxU <= parc->bandwidth)
					{
						scoreTable[u] = -1; // indicating this network node must not be deployed with a server
						break;
					}
					// TODO: more stategies.
				}
				else // decide which access node to deploy a server or both not deployed with server
				{
					uint32_t consumptionIdx = accessNodeMap[v];
					int demandFluxV = demandTable[consumptionIdx];
					scanned[consumptionIdx] = true;
					// calculate if one of them have to deploy a server
					int fluxInputCapTotal = 0, neighborAccessNumU = 0, neighborAccessNumV = 0;
					for (qarc = graph.NodeList[v].firstArc; qarc != NULL; qarc = qarc->nextArc)
					{
						if (qarc->dstID != u && qarc->dstID < mNetworkNodeNum) // ensure qarc points to an external network node
							fluxInputCapTotal += qarc->bandwidth;
						if (qarc->dstID != u && accessNodeMap.find(qarc->dstID) != accessNodeMap.end())
							++neighborAccessNumV;
					}
					for (qarc = graph.NodeList[u].firstArc; qarc != NULL; qarc = qarc->nextArc)
					{
						if (qarc->dstID != v && qarc->dstID < mNetworkNodeNum) // ensure qarc points to an external network node
							fluxInputCapTotal += qarc->bandwidth;
						if (qarc->dstID != v && accessNodeMap.find(qarc->dstID) != accessNodeMap.end())
							++neighborAccessNumU;
					}
					debug_log("fluxInputCapTotal of [%d] and [%d]: %d\n", u, v, fluxInputCapTotal);
					if (fluxInputCapTotal < demandFluxU + demandFluxV) // one of them have to deploy a server
					{
						// FIXME: this strategy is too simple
						if (neighborAccessNumU > neighborAccessNumV)
							scoreTable[u] = INF;
						else if (neighborAccessNumU < neighborAccessNumV)
							scoreTable[v] = INF;
						else
						{
							if (demandFluxU > demandFluxV)
								scoreTable[u] = INF;
							else if (demandFluxU < demandFluxV)
								scoreTable[v] = INF;
							else if (outLinkFee[u] >= outLinkFee[v])
								scoreTable[u] = INF;
							else
								scoreTable[v] = INF;
						}
						// TODO: calculate the situation when server deploys on current access node
						// TODO: calculate the situation when server deploys on access node v

						// print log message
						if (scoreTable[u] == INF)
							info_log("access node [%d] have to deploy a server..\n", u);
						else
							info_log("access node [%d] have to deploy a server..\n", v);
					}
					else
					{
						// TODO: hard to decide
					}
				}
			}
			else if (v < mNetworkNodeNum) // when this neighbor is not an access node, ensure it is not a consumption node
			{
				// FIXME: this strategy is too static
				curLinkFlux = demandFluxU >= parc->bandwidth ? parc->bandwidth : demandFluxU;
				scoreTable[v] += curLinkFlux * (10 - parc->rental);
			}
			parc = parc->nextArc;
		}
		
	}
}

void TopologyScanner::_decideMultiplicativeFactor()
{
	switch (cspDensity)
	{
	case CspDensity::UltraSparse:
		multiplicativeFactor = 4.0; break;
	case CspDensity::VerySparse:
		multiplicativeFactor = 3.0; break;
	case CspDensity::Sparse:
		multiplicativeFactor = 2.5; break;
	case CspDensity::General:
		multiplicativeFactor = 2.0; break;
	case CspDensity::Dense:
		multiplicativeFactor = 1.0; break;
	case CspDensity::VeryDense:
		multiplicativeFactor = 0.8; break;
	case CspDensity::UltraDense:
		multiplicativeFactor = 0.6; break;
	default:
		multiplicativeFactor = 2.0; // the same as general case
	}
}

void TopologyScanner::_selectByScore()
{
	_decideMultiplicativeFactor();

	std::vector<std::pair<int /* score */, int /* ID */> > scoreIDPair;

	uint32_t i = 0, candidateNum = static_cast<uint32_t>(multiplicativeFactor*consumptionNodeNum);
	for (i = 0; i < networkNodeNum; ++i)
		scoreIDPair.emplace_back(scoreTable[i], i);

	std::sort(scoreIDPair.begin(), scoreIDPair.end(), std::greater<std::pair<int, int> >());

	debug_log("score sort from high to low: ");
	for (i = 0; i < candidateNum; ++i)
	{
		if (log_level >= DEBUG_LEVEL)
			printf("%d ", scoreIDPair[i].first);
		possibleServers.push_back(scoreIDPair[i].second);
	}
	printf("\n");
}
