/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: TopologyScanner.h                                                          *
 * @author: Xu Le                                                                    *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#ifndef __TOPOLOGYSCANNER_H__
#define __TOPOLOGYSCANNER_H__

#include "Graph.h"

/** A class undertaking to scan graph topology to find some useful results. */
class TopologyScanner
{
public:
	explicit TopologyScanner(Graph& _graph);
	~TopologyScanner();

	TopologyScanner(const TopologyScanner& rhs) = delete;
	TopologyScanner& operator=(const TopologyScanner& rhs) = delete;

	void scan();
	void recommend(std::list<std::vector<int> >& initialServerGroups);

	/** API for Greedy. */
	int* getOutLinkCap() { return outLinkCap; }
	/** API for Greedy. */
	double* getRentalDeployFeeRatio() { return rentalDeployFeeRatio; }

private:
	/** description of the density of arcs in graph. */
	enum class ArcDensity {
		Unknown = 0,
		UltraSparse,
		VerySparse,
		Sparse,
		General,
		Dense,
		VeryDense,
		UltraDense
	};
	/** description of the density of consumption nodes in graph. */
	enum class CspDensity {
		Unknown = 8,
		UltraSparse,
		VerySparse,
		Sparse,
		General,
		Dense,
		VeryDense,
		UltraDense
	};
	const char* _arcDensityString();
	const char* _cspDensityString();
	void _defineDensity();
	void _mustDeployOnAccess();
	void _calcCheapestAmongAccess();
	void _scanNeighborOfAccess();
	void _decideMultiplicativeFactor();
	void _selectByScore();

private:
	Graph &graph; ///< reference to graph.
	ArcDensity arcDensity; ///< stores the density of arcs in graph.
	CspDensity cspDensity; ///< stores the density of consumption nodes in graph.
	int mNetworkNodeNum;      ///< equals to networkNodeNum, just get rid of annoying type incompatible warning.
	int mServerDeploymentFee; ///< equals to serverDeploymentFee, just get rid of annoying type incompatible warning.
	int lowerBoundNum;  ///< lower bound number of network nodes needed to deploy server.
	int upBoundNum;     ///< up bound number of network nodes needed to deploy server.

	/** @name status of all network nodes. */
	///@{
	int *outLinkNum;    ///< the number of links of a network node.
	int *outLinkCap;    ///< the sum up capacity of all links of a network node.
	double *outLinkFee; ///< the weighted average rental fee of all links of a network node.
	double *capRank;    ///< the sum up capacity rank of a network node.
	double *feeRank;    ///< the weighted average rental fee rank of a network node.
	int *scoreTable;    ///< stores score of all network nodes.
	///@}

	/** @name status of access network nodes. */
	///@{
	double *rentalDeployFeeRatio; ///< when an access node doesn't deploy server, it will need a minimum link rental fee X, this ratio value certainly is X/singleServerFee.
	///@}

	double multiplicativeFactor;  ///< the result multiplied by consumption node num is the number of network nodes to be selected as possible servers' deployment location. (decided by cspDensity)
	std::vector<int> possibleServers; ///< all network nodes with high score that are possibly chosen as where to deploy servers.
};

#endif /* __TOPOLOGYSCANNER_H__ */
