/*************************************************************************************
 * @copyright: Huawei Technologies Co. Ltd.                                          *
 * @file: main.cpp                                                                   *
 * @author: Xu Le, He Jianfan                                                        *
 * This file is part of 2017 Huawei CodeCraft solution <http://codecraft.huawei.com> *
 *************************************************************************************/
#include "Greedy.h"

extern int log_level;

extern std::vector<int> accessTable;
extern std::vector<int> demandTable;

uint32_t networkNodeNum = 0;
uint32_t linkNum = 0;
uint32_t consumptionNodeNum = 0;
uint32_t serverDeploymentFee = 0;

std::list<int> bestServers;

int main(int argc, char *argv[])
{
	Timer procedureTimer("Elapsed Time: ");
	/*******************************************************************
	 ********************     Time elapsing ...     ********************
	 *******************************************************************/
	printf("current log level: %s\n\n", getLogLevel());

	std::list<struct LinkTuple> linkTuples;

	// 1. ===============    parse input char* array    ===============
	parseInput(argv[1], linkTuples);

	debug_log("NetworkNodeNum: %u, LinkNum: %u, ConsumptionNodeNum: %u\n", networkNodeNum, linkNum, consumptionNodeNum);
	debug_log("ServerDeploymentFee: %u\n", serverDeploymentFee);

	{
		Complexity defaultComplexity = defineDefaultComplexity(); // local variable, just print the complexity for debug, not used to optimize case according to its size
		info_log("default complexity: %s\n", complexityString(defaultComplexity));
	}

	linkTuples.sort([](const struct LinkTuple& lhs, const struct LinkTuple& rhs) { return lhs.src < rhs.src; });

	// 2. ===============    construct graph    ===============
	Graph graph(networkNodeNum + consumptionNodeNum);
	graph.construct(linkTuples);
	graph.undirectify(linkTuples);
	linkTuples.clear(); // we don't need it anymore, release its memory

	// 3. ===============    display graph    ===============
	// graph.display(); // log control inside

	// 4. ===============    add minimum cost flow algorithm related components into graph    ===============
	graph.preferZKW(true);
	graph.preMinimumCostMaximumFlow();

	// 5. ===============    critical part of finding a server deployment solution    ===============
	std::list<int> newServers;
	std::list<std::vector<int> > initialServerCombinations;

	// 5.1. ==========    scan topology to obtain combinations of some carefully selected network nodes as initial scheme    ==========
	TopologyScanner topoScanner(graph);

	topoScanner.scan();
	topoScanner.recommend(initialServerCombinations);

	// 5.2. ==========    deploy CDN servers on one of these combinations of network nodes    ==========
	Greedy greedy(graph, initialServerCombinations);
	greedy.obtainOutLinkCap(topoScanner.getOutLinkCap());
	greedy.obtainRentalDeployFeeRatio(topoScanner.getRentalDeployFeeRatio());
	greedy.initialize(newServers);

	// info_log("initial servers: ");
	// printList(newServers, "");

	int loop = 1;
	bool continueLoop = true;
	double elapsedTime = 0.0;
	long minRentalDeploymentFee = INF;
	while (true)
	{
		// 5.3. ==========    add a super source onto graph    ==========
		graph.addSuperSource(newServers);

		// 5.4. ==========    execute minimum cost flow algorithm    ==========
		long rentalFee = graph.minimumCostMaximumFlow(newServers);
		long deploymentFee = serverDeploymentFee * static_cast<long>(newServers.size());

		// info_log("Link fee is %ld, server fee is %ld, sum up %ld.\n", rentalFee, deploymentFee, rentalFee + deploymentFee);

		// 5.5. ==========    a better deployment solution ?    ==========
		if (minRentalDeploymentFee > rentalFee + deploymentFee && rentalFee != -1)
		{
			minRentalDeploymentFee = rentalFee + deploymentFee;
			bestServers.assign(newServers.begin(), newServers.end());
		}

		continueLoop = greedy.optimize(newServers, rentalFee);

		// 5.6. ==========    time is not enough    ==========
		if (loop % 50 == 0)
			elapsedTime = procedureTimer.elapsed();
		if (loop % 1000 == 0)
			info_log("Time elapsed: %fs\n", elapsedTime);
		// how many loops do you want?
		if (++loop > 10000 || !continueLoop || elapsedTime > 87.0) // we have to exit loop when only 3 seconds left
		{
			printf("INFO | Exit loop: %d\n", loop);
			info_log("best servers: ");
			printList(bestServers, "");

			// 5.6.1. =====    delete current super source from graph and revert the capacity of arc to initialization    =====
			graph.revertStatusToInitialization();
			graph.delSuperSource();

			// 5.6.2. =====    add the super source of best servers' union onto graph    =====
			graph.addSuperSource(bestServers);

			// 5.6.3. =====    calculate minimum cost flow using best servers and then record corresponding flow path    =====
			graph.minimumCostMaximumFlow(bestServers);
			graph.recordFlowPath();

			// 5.6.4. =====    in fact, it is not must to revert capacity and delete super source    =====
			graph.revertStatusToInitialization();
			graph.delSuperSource();
			break;
		}

		// 5.7. ==========    delete the super source from graph and revert the capacity of arc to initialization    ==========
		graph.revertStatusToInitialization();
		graph.delSuperSource();

		// info_log("new servers: ");
		// printList(newServers, "");
	}

	graph.postMinimumCostMaximumFlow();

	info_log("Minimum link rental fee plus server deployment fee is %ld.\n", minRentalDeploymentFee);

	for (std::list<std::vector<int> >::iterator itfp = graph.fluxPathList.begin(); itfp != graph.fluxPathList.end(); ++itfp)
	{
		std::vector<int>& fluxPath = *itfp;
		fluxPath[fluxPath.size()-2] -= static_cast<int>(networkNodeNum);
		// printVector(fluxPath, "");
	}

	// 6. ===============    verify server deployment and flux paths scheme    ===============
	long totalRentalFee = 0;
	bool schemeOK = verifyScheme(graph, graph.fluxPathList, totalRentalFee);
	uint32_t totalDeploymentFee = serverDeploymentFee * static_cast<uint32_t>(bestServers.size());
	if (schemeOK == true)
		uncond_log("Congratulations! Total link rental fee is %ld, deployment fee is %u, sum up %ld.\n", totalRentalFee, totalDeploymentFee, totalRentalFee + totalDeploymentFee);
	else
		error_log("Shit! The scheme failed to pass verification!\n");
	
	// 7. ===============    destruct graph    ===============
	graph.destruct();

	// 8. ===============    output flux path list to result file    ===============
	outputToFile(argv[2], graph.fluxPathList);

	/*******************************************************************
	 ********************     Time elapsing end     ********************
	 *******************************************************************/
}
